# Openvpn helper things
from django.conf import settings
from django.contrib.auth.models import User

from gamoto import users

import os
import ipaddress


def _readlog(statuslog):
    clients = {}
    with open(statuslog) as log:
        current_stat = None

        for ln in log:
            if ln.startswith('Common Name'):
                current_stat = 'clients'

            elif ln.startswith('Virtual Address'):
                current_stat = 'routes'

            elif ln.startswith('GLOBAL') or ln.startswith('ROUTING'):
                current_stat = None

            elif current_stat == 'clients':
                name, remote_ip, rx, tx, start = ln.strip().split(',')

                clients[name] = {
                    'ip': remote_ip.split(':')[0],
                    'rx_bytes': int(rx),
                    'tx_bytes': int(tx),
                    'connected': start
                }

            elif current_stat == 'routes':
                virtual, name, remote, start = ln.strip().split(',')
                clients[name]['virtual'] = virtual
    return clients


def getStatus():
    statuslog = os.path.join(settings.BASE_PATH, 'openvpn-status.log')

    if not os.path.exists(statuslog):
        return {}

    try:
        clients = _readlog(statuslog)
    except PermissionError as err:
        # Try to fix it
        from gamoto import users
        users.sudo(
            'chown',
            '%s:%s' % (settings.GAMOTO_USER, settings.GAMOTO_GROUP),
            statuslog
        )
        clients = _readlog(statuslog)

    return clients


def getClient(name):
    return getStatus().get(name, None)


def getRoutes(user):
    if isinstance(user, str):
        user = User.objects.get(username=user)

    pwd = users.getUser(user.username)

    if not (pwd and pwd.get('mfa')):
        return None

    subnets = []
    groups = user.groups.all()

    for group in groups:
        permissions = group.permissions.all()
        for p in permissions:
            subnet = p.codename.split('_')[-1]

            if subnet not in subnets:
                subnets.append(subnet)

    return subnets


def updateCCDs():
    db_users = User.objects.all()

    ccd_path = os.path.join(settings.BASE_PATH, 'ccd')
    if not os.path.exists(ccd_path):
        os.makedirs(ccd_path)

    for user in db_users:
        subnets = getRoutes(user)
        if subnets:
            user_ccd_path = os.path.join(ccd_path, user.username)
            with open(user_ccd_path, 'wt') as user_ccd:
                for subnet in subnets:
                    net = ipaddress.ip_network(subnet)
                    user_ccd.write('push "route %s %s"\n' % (
                        net.network_address.exploded,
                        net.netmask.exploded
                    ))



"""gamoto URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/1.11/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  url(r'^$', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  url(r'^$', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.conf.urls import url, include
    2. Add a URL to urlpatterns:  url(r'^blog/', include('blog.urls'))
"""
from django.contrib import admin
from django.urls import path, include
from django.contrib.auth import views as auth_views
from django.conf import settings

from gamoto import views


urlpatterns = [
    path('admin/', admin.site.urls),
    path('', include('social_django.urls', namespace='social')),
    path(
        'accounts/login/',
        auth_views.LoginView.as_view(
            template_name='login.html',
            extra_context={'oauth': settings.GOOGLE_AUTH}
        ),
        name="login"
    ),
    path('accounts/', include('django.contrib.auth.urls')),
    path('', views.index, name="index"),
    path('enroll', views.enroll_user, name="enroll"),
    path('reset_2fa', views.reset_2fa, name="reset_2fa"),
    path('vpn_zip', views.vpn_zip, name="vpn_zip"),
    path('endpoints', views.admin_endpoints, name="endpoints"),
    path('create_group', views.group_create, name="create_group"),
    path('delete_group/<int:group_id>/', views.group_delete,
         name="delete_group"),
    path('create_subnet', views.subnet_create, name="create_subnet"),
    path('add_group_subnet/<int:group_id>/', views.group_subnet_add,
         name="add_group_subnet"),
    path('remove_group_subnet/<int:group_id>/<int:permission_id>/',
         views.group_subnet_remove, name="remove_group_subnet"),
    path('user_groups/<int:user_id>/',
         views.user_group_modify, name="user_groups"),
    path('users', views.admin_users, name="users")
]



from django.core.management.base import BaseCommand

from django.conf import settings
from django.contrib.auth.models import User

from gamoto import users, openvpn

import os
import sys


class Command(BaseCommand):
    help = 'OpenVPN client-connect scripts'

    def add_arguments(self, parser):
        parser.add_argument('client_cmd_path', nargs='?', type=str)

    def getIptables(self):
        result, err = users.sudo('iptables-save')
        result = result.strip()
        rules = [i for i in result.decode().split('\n')
                 if not i.startswith('#')]

        tables = {}
        table = None
        for r in rules:
            if r.startswith('*'):
                table = r.strip('*')
                tables[table] = {}
            elif r.startswith(':'):
                chain, default, b = r.strip(':').split()
                if default == '-':
                    default = None

                tables[table][chain] = {
                    'default': default,
                    'rules': []
                }
            elif r.startswith('-'):
                token, chain, args = r.split(None, 2)
                tables[table][chain]['rules'].append(args)
        return tables

    def iptables(self, rule):
        result, err = users.sudo('/sbin/iptables', *rule.split())
        if err:
            raise Exception(err)

    def setupIptables(self):
        # Makes sure iptables has our chain
        ipt = self.getIptables()

        if 'openvpn' not in ipt['filter']:
            self.iptables('-N openvpn')

        jump_rule = '-i %s -j openvpn' % settings.VPN_INTERFACE

        if jump_rule not in ipt['filter']['INPUT']['rules']:
            self.iptables('-I INPUT 1 -i tun0 -j openvpn')

    def allowClient(self, name, src, dest):
        if not (name and src and dest):
            return None

        rules = self.getIptables()['filter']['openvpn']['rules']

        rule = '-A openvpn -i tun0 -s %s -d %s -m comment --comment "%s"' % (
            src, dest, name
        )

        if rule not in rules:
            self.iptables(rule)

    def flushClient(self, name):
        rules = self.getIptables()['filter']['openvpn']['rules']
        for i, r in enumerate(rules):
            idx = str(i + 1)
            if name in r.split()[-1]:
                users.sudo('/sbin/iptables', '-D', 'openvpn', idx)

    def connect(self, user):
        subnets = openvpn.getRoutes(user)
        ip = os.getenv('ifconfig_pool_remote_ip')
        self.setupIptables()

        for subnet in subnets:
            self.allowClient(user, ip, subnet)

    def disconnect(self, user):
        self.flushClient(user)

    def handle(self, *args, **options):
        script_type = os.getenv('script_type')
        user = os.getenv('common_name')
        if not user:
            self.stderr.write(
                'Error, no CN found. This script must be run by OpenVPN!')
            sys.exit(1)

        if settings.MANAGE_IPTABLES:
            if script_type == 'client-connect':
                self.connect(user)

            elif script_type == 'client-disconnect':
                self.disconnect(user)

            else:
                self.stderr.write('Not sure what you want me to do :(')
                sys.exit(1)



"""
Django settings for gamoto project.
"""
import configparser
import os


GAMOTO_USER = 'gamoto'
GAMOTO_GROUP = 'gamoto'

BASE_PATH = '/var/lib/gamoto'
CA_PATH = os.path.join(BASE_PATH, 'ca')
USER_PATH = os.path.join(BASE_PATH, 'users')

OPENVPN_PORT = 1194
OPENVPN_HOSTNAME = 'vpn.acme.zp'

VPN_INTERFACE = 'tun0'

MANAGE_IPTABLES = True

CA_SETUP = {
    'org': 'ACME Corp',
    'ou': 'IT',
    'email': 'support@acme.zp',
    'city': 'ACME Village',
    'state': 'Assur',
    'country': 'US'
}

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

SECRET_KEY = '%#*6c91zihs)_ptl1t3sxu$$#a)bah)wv8$-i6m*646x_gsl+n'

DEBUG = False

ALLOWED_HOSTS = ['*']

LOGOUT_REDIRECT_URL = "/"
LOGIN_REDIRECT_URL = "/"

AUTHENTICATION_BACKENDS = (
    'social_core.backends.google.GoogleOAuth2',
    'django.contrib.auth.backends.ModelBackend',
)

GOOGLE_AUTH = False
SOCIAL_AUTH_GOOGLE_OAUTH2_KEY = ''
SOCIAL_AUTH_GOOGLE_OAUTH2_SECRET = ''

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'social_django',
    'gamoto'
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'gamoto.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [os.path.join(BASE_DIR, "gamoto/templates")],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = 'gamoto.wsgi.application'


AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation'
                '.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation'
                '.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation'
                '.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation'
                '.NumericPasswordValidator',
    },
]

LANGUAGE_CODE = 'en-us'
TIME_ZONE = 'UTC'
USE_I18N = True
USE_L10N = True
USE_TZ = True

STATIC_URL = '/static/'
STATIC_ROOT = os.path.join(BASE_PATH, "static")

# Read system config

config = configparser.ConfigParser()
config.read('/etc/gamoto/gamoto.conf')
# Witchcraft
if 'main' in config:
    # Get MFA setting
    MFA_ENABLED = config['main'].getboolean('2fa', True)

    GAMOTO_USER = config['main'].get('user', GAMOTO_USER)
    GAMOTO_GROUP = config['main'].get('group', GAMOTO_GROUP)

    # Default paths
    BASE_PATH = config['main'].get('path', BASE_PATH)
    CA_PATH = config['main'].get('ca_path', os.path.join(BASE_PATH, 'ca'))
    USER_PATH = config['main'].get('home_path',
                                   os.path.join(BASE_PATH, 'users'))

    STATIC_PATH = config['main'].get(
        'static_path', os.path.join(BASE_PATH, 'static'))

    SECRET_KEY = config['main'].get('cookie_secret', SECRET_KEY)
    MANAGE_IPTABLES = config['main'].getboolean('manage_iptables',
                                                MANAGE_IPTABLES)

    DEBUG = config['main'].getboolean('debug', DEBUG)

if 'google' in config:
    GOOGLE_AUTH = True
    SOCIAL_AUTH_GOOGLE_OAUTH2_KEY = config['google'].get('key', None)
    SOCIAL_AUTH_GOOGLE_OAUTH2_SECRET = config['google'].get('secret', None)
    whitelisted = config['google'].get('allowed_domains', None)
    if whitelisted:
        whitelisted = whitelisted.replace(' ', '').split(',')
        SOCIAL_AUTH_GOOGLE_OAUTH2_WHITELISTED_DOMAINS = whitelisted

if 'openvpn' in config:
    # Openvpn settings
    OPENVPN_PORT = config['openvpn'].get('port', OPENVPN_PORT)
    OPENVPN_HOSTNAME = config['openvpn'].get('hostname', OPENVPN_HOSTNAME)

if 'ca' in config:
    # Merge CA config
    for key in CA_SETUP.keys():
        CA_SETUP[key] = config['ca'].get(key, CA_SETUP[key])

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_PATH, 'gamoto.db'),
    }
}



from django.core.management.base import BaseCommand, CommandError
from django.db.utils import IntegrityError
from django.conf import settings
# from django.contrib.auth.models import User, Group, Permission
from django.contrib.contenttypes.models import ContentType

from gamoto import ca, users

import os
import sys


class Command(BaseCommand):
    help = 'Configure the CA'

    def handle(self, *args, **options):
        def _status(var):
            if var:
                self.stdout.write(self.style.SUCCESS('OK'))
            else:
                self.stdout.write(self.style.SUCCESS('EXISTS'))

        if (users.getSystemUID() != os.getuid()):
            self.stdout.write(self.style.ERROR(
                "Please run this command as root, or " + settings.GAMOTO_USER))
            sys.exit(1)

        # Ensure paths exist
        for path in [settings.BASE_PATH, settings.CA_PATH, settings.USER_PATH]:
            if not os.path.isdir(path):
                os.makedirs(path)

        # Setup CA
        myca = ca.CertificateAuthority(
            settings.CA_SETUP['ou'],
            settings.CA_SETUP['org'],
            settings.CA_SETUP['email'],
            settings.CA_SETUP['country'],
            settings.CA_SETUP['state'],
            settings.CA_SETUP['city'],
            key_path=settings.CA_PATH
        )

        self.stdout.write("Creating Diffie-Hellman parameters... ", ending="")
        self.stdout.flush()
        _status(myca.createDH())

        self.stdout.write("Building certificate authority... ", ending="")
        self.stdout.flush()
        _status(myca.createCA('openvpn'))

        self.stdout.write("Creating server certificate request... ", ending="")
        self.stdout.flush()
        _status(myca.createCSR('openvpn'))

        self.stdout.write("Signing server certificate... ", ending="")
        self.stdout.flush()
        _status(myca.signCSR('openvpn', server=True))

        self.stdout.write("Generating CRL... ", ending="")
        self.stdout.flush()
        _status(myca.createCRL())

        # Setup DB
        self.stdout.write("Creating types and permissions... ", ending="")
        self.stdout.flush()
        try:
            ctype = ContentType.objects.create(
                app_label='subnet',
                model='group'
            )
            ctype.save()
            _status(True)
        except IntegrityError:
            _status(False)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.http import HttpResponse
from django.shortcuts import render, redirect
from django.contrib.auth import views as auth_views
from django.contrib.auth.models import User, Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.contrib.auth.decorators import login_required, user_passes_test
from django.core.exceptions import ObjectDoesNotExist

from gamoto import users, openvpn, forms


def returnFile(name, content_type, data):
    response = HttpResponse(data, content_type=content_type)
    response['Content-Disposition'] = 'attachment; filename="%s"' % name
    return response


@login_required
def index(request):
    user_name = request.user.username

    passwd = users.getUser(user_name)

    vpn = openvpn.getClient(user_name)

    return render(request, "index.html", {
        'passwd': passwd,
        'vpn': vpn,
        'admin': request.user.is_superuser,
        'name': request.user.get_full_name(),
        'sbactive': 'index'
    })


@login_required
def vpn_zip(request):
    user_name = request.user.username
    passwd = users.getUser(user_name)

    return returnFile(user_name + ".zip", "application/zip",
                      users.getVPNZIP(user_name))


@login_required
def reset_2fa(request):
    user_name = request.user.username
    codes, authurl = users.configureTOTP(user_name)

    return render(request, "enroll.html", {
        'name': request.user.get_full_name(),
        'authurl': authurl,
        'codes': codes,
        'reset': True
    })


@login_required
def enroll_user(request):
    user_name = request.user.username
    passwd = users.getUser(user_name)

    if passwd:
        return redirect('index')

    users.createUser(user_name)
    users.createVPN(user_name)
    codes, authurl = users.configureTOTP(user_name)
    openvpn.updateCCDs()

    return render(request, "enroll.html", {
        'name': request.user.get_full_name(),
        'authurl': authurl,
        'codes': codes
    })


@user_passes_test(lambda u: u.is_superuser)
def admin_endpoints(request):
    groups = []

    for group in Group.objects.all():
        subnets = [{
            'id': perm.id,
            'name': perm.name,
            'subnet': perm.codename.split('_')[-1]
        } for perm in group.permissions.all()]

        groups.append({
            'id': group.id,
            'name': group.name,
            'subnets': subnets
        })

    return render(request, "admin_endpoints.html", {
        'sbactive': 'endpoints',
        'admin': request.user.is_superuser,
        'groups': groups
    })


@user_passes_test(lambda u: u.is_superuser)
def group_create(request):
    if request.method == "POST":
        form = forms.GroupForm(request.POST)
        if form.is_valid():
            group = form.save(commit=False)
            group.save()
            form.save_m2m()
            openvpn.updateCCDs()

            return redirect('endpoints')
    else:
        form = forms.GroupForm()

    return render(request, 'admin_group_create.html', {
        'form': form,
        'sbactive': 'endpoints',
        'admin': request.user.is_superuser
    })


@user_passes_test(lambda u: u.is_superuser)
def subnet_create(request):
    if request.method == "POST":
        form = forms.SubnetForm(request.POST)
        if form.is_valid():
            subnet = form.save(commit=False)
            subnet.save()
            openvpn.updateCCDs()

            return redirect('endpoints')
    else:
        form = forms.SubnetForm()

    return render(request, 'admin_subnet_create.html', {
        'form': form,
        'sbactive': 'endpoints',
        'admin': request.user.is_superuser
    })


@user_passes_test(lambda u: u.is_superuser)
def group_delete(request, group_id):
    grp = Group.objects.get(id=group_id)
    grp.delete()
    openvpn.updateCCDs()
    return redirect('endpoints')


@user_passes_test(lambda u: u.is_superuser)
def group_subnet_remove(request, group_id, permission_id):
    grp = Group.objects.get(id=group_id)
    perm = Permission.objects.get(id=permission_id)

    grp.permissions.remove(perm)

    openvpn.updateCCDs()

    return redirect('endpoints')


@user_passes_test(lambda u: u.is_superuser)
def group_subnet_add(request, group_id):
    group = Group.objects.get(id=group_id)

    if request.method == "POST":
        form = forms.GroupSubnetForm(request.POST)
        if form.is_valid():
            gslink = form.cleaned_data

            perm_name = "network_%s" % gslink['subnet']

            content_type = ContentType.objects.get(app_label='subnet')

            try:
                permission = Permission.objects.get(codename=perm_name)
            except ObjectDoesNotExist:
                permission = Permission.objects.create(
                    codename=perm_name,
                    name=gslink['name'],
                    content_type=content_type
                )
                permission.save()

            group.permissions.add(permission)
            openvpn.updateCCDs()

            return redirect('endpoints')
    else:
        form = forms.GroupSubnetForm()

    return render(request, 'admin_subnet_add.html', {
        'form': form,
        'sbactive': 'endpoints',
        'group_id': group_id,
        'group': group,
        'admin': request.user.is_superuser
    })


@user_passes_test(lambda u: u.is_superuser)
def admin_users(request):
    all_users = []
    for user in User.objects.all():
        user_name = user.username
        try:
            social = user.social_auth.get()
        except ObjectDoesNotExist as err:
            continue

        groups = user.groups.all()
        passwd = users.getUser(user_name)

        enrolled = False
        if passwd:
            enrolled = True

        vpn = openvpn.getClient(user_name)

        all_users.append({
            'username': user_name,
            'id': user.id,
            'name': user.get_full_name(),
            'email': user.email,
            'provider': social.provider,
            'last_login': user.last_login,
            'enrolled': enrolled,
            'vpn': vpn,
            'groups': [g.name for g in groups]
        })

    all_users.sort(key=lambda x: x['username'])

    return render(request, "admin_users.html", {
        'sbactive': 'users',
        'users': all_users,
        'admin': request.user.is_superuser
    })


@user_passes_test(lambda u: u.is_superuser)
def user_group_modify(request, user_id):
    user = User.objects.get(id=user_id)

    if request.method == "POST":
        form = forms.UserGroupForm(request.POST, instance=user)
        if form.is_valid():
            user_f = form.save(commit=False)
            user_f.save()
            form.save_m2m()

            openvpn.updateCCDs()

            return redirect('users')
    else:
        form = forms.UserGroupForm(instance=user)

    return render(request, 'user_group_modify.html', {
        'form': form,
        'sbactive': 'endpoints',
        'user_id': user.id,
        'admin': request.user.is_superuser,
        'name': user.get_full_name()
    })



# CA management
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa, dh
from cryptography.x509.oid import NameOID, ExtendedKeyUsageOID

import datetime
import uuid
import os


class CertificateAuthority(object):
    """
    Certificate authority class
    """
    def __init__(self, ou, org, email, country, province, city, key_path=None,
                 validity=3653):
        """
        Construct a CertificateAuthority object
        parameters:
            org: Organisation name
            ou: Organisation unit
            email: Email address
            country: Country
            province: Province
            city: City
            key_path (kw): Path to store and retrieve PKI
                           default - Current working directory
            validity (kw): Certificate validity in days
                           default - 3653 (~10 years)
        """
        if key_path:
            self.key_path = key_path
        else:
            self.key_path = os.path.dirname(os.path.dirname(
                os.path.abspath(__file__)))

        if not os.path.exists(self.key_path):
            raise Exception("Key path does not exist")

        self.x509params = [
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, ou),
            x509.NameAttribute(NameOID.ORGANIZATIONAL_UNIT_NAME, org),
            x509.NameAttribute(NameOID.COUNTRY_NAME, country),
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, province),
            x509.NameAttribute(NameOID.LOCALITY_NAME, city),
            x509.NameAttribute(NameOID.EMAIL_ADDRESS, email)
        ]

        self.ca_key_path = os.path.join(self.key_path, "ca.key")
        self.ca_cert_path = os.path.join(self.key_path, "ca.crt")
        self.crl_path = os.path.join(self.key_path, "crl.pem")

        self.validity_time = datetime.timedelta(validity, 0, 0)

    def createDH(self, key_size=2048):
        """
        Generate Diffie-Hellman key
        """
        filename = os.path.join(self.key_path, "dh%s.pem" % key_size)
        if os.path.exists(filename):
            return False

        parameters = dh.generate_parameters(generator=2, key_size=key_size,
                                            backend=default_backend())
        dhbytes = parameters.parameter_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.ParameterFormat.PKCS3
        )

        with open(filename, "wb") as f:
            f.write(dhbytes)

        return True

    def getCA(self):
        """
        Load the default CA keys
        """
        if (os.path.exists(self.ca_key_path) and os.path.exists(
                self.ca_cert_path)):
            with open(self.ca_cert_path, "rb") as cafile:
                ca_cert = x509.load_pem_x509_certificate(cafile.read(),
                                                         default_backend())

            with open(self.ca_key_path, "rb") as keyfile:
                private_key = serialization.load_pem_private_key(
                    keyfile.read(),
                    password=None,
                    backend=default_backend()
                )

            return (ca_cert, private_key)
        else:
            raise Exception("CA does not exist")

    def createCA(self, cn):
        """
        Create a CA with name cn
        """
        if (os.path.exists(self.ca_key_path) and os.path.exists(
                self.ca_cert_path)):
            return False

        private_key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend()
        )

        public_key = private_key.public_key()

        common_name = x509.NameAttribute(NameOID.COMMON_NAME, cn)
        now = datetime.datetime.utcnow()

        sn = x509.Name(self.x509params + [common_name])

        builder = x509.CertificateBuilder().subject_name(
            sn
        ).issuer_name(
            sn
        ).not_valid_before(
            now - datetime.timedelta(0, 1, 0)
        ).not_valid_after(
            now + self.validity_time
        ).serial_number(
            int(uuid.uuid4())
        ).public_key(
            public_key
        ).add_extension(
            x509.BasicConstraints(ca=True, path_length=None),
            critical=True
        ).add_extension(
            x509.SubjectKeyIdentifier.from_public_key(
                private_key.public_key()),
            critical=False
        ).add_extension(
            x509.AuthorityKeyIdentifier.from_issuer_public_key(
                private_key.public_key()),
            critical=False
        )

        certificate = builder.sign(
            private_key=private_key, algorithm=hashes.SHA256(),
            backend=default_backend()
        )

        with open(self.ca_key_path, "wb") as f:
            f.write(private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption()
            ))

        with open(self.ca_cert_path, "wb") as f:
            f.write(certificate.public_bytes(
                encoding=serialization.Encoding.PEM,
            ))

        return True

    def getCSR(self, cn):
        """
        Retrieve a CSR with name cn
        """
        csr_path = os.path.join(self.key_path, "%s.csr" % cn)

        if os.path.exists(csr_path):
            with open(csr_path, "rb") as csrfile:
                csr = x509.load_pem_x509_csr(csrfile.read(), default_backend())
            return csr
        else:
            raise Exception("CSR does not exist")

    def createCSR(self, cn):
        """
        Create a CSR with name cn
        """
        key_path = os.path.join(self.key_path, "%s.key" % cn)
        csr_path = os.path.join(self.key_path, "%s.csr" % cn)

        if (os.path.exists(key_path) and os.path.exists(csr_path)):
            return False

        private_key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend()
        )

        common_name = x509.NameAttribute(NameOID.COMMON_NAME, cn)

        builder = x509.CertificateSigningRequestBuilder().subject_name(
            x509.Name(self.x509params + [common_name])
        ).add_extension(
            x509.SubjectAlternativeName([
                x509.DNSName(cn),
            ]),
            critical=False,
        ).sign(private_key, hashes.SHA256(), default_backend())

        # Write key
        with open(key_path, "wb") as f:
            f.write(private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption()
            ))

        # Write CSR
        with open(csr_path, "wb") as f:
            f.write(builder.public_bytes(serialization.Encoding.PEM))

        return True

    def signCSR(self, cn, server=False):
        """
        Sign a CSR of name cn. Set server kw to True if this is for TLS
        server validation
        """
        assert isinstance(server, bool)

        cert_path = os.path.join(self.key_path, "%s.crt" % cn)

        if os.path.exists(cert_path):
            return False

        ca_cert, private_key = self.getCA()
        csr = self.getCSR(cn)

        now = datetime.datetime.utcnow()
        if server:
            extension = ExtendedKeyUsageOID.SERVER_AUTH
        else:
            extension = ExtendedKeyUsageOID.CLIENT_AUTH

        certificate = x509.CertificateBuilder().subject_name(
            csr.subject
        ).issuer_name(
            ca_cert.subject
        ).public_key(
            csr.public_key()
        ).serial_number(
            int(uuid.uuid4())
        ).not_valid_before(
            now - datetime.timedelta(0, 1, 0)
        ).not_valid_after(
            now + self.validity_time
        ).add_extension(
            x509.BasicConstraints(ca=False, path_length=None),
            critical=False
        ).add_extension(
            x509.SubjectKeyIdentifier.from_public_key(
                private_key.public_key()),
            critical=False
        ).add_extension(
            x509.AuthorityKeyIdentifier.from_issuer_public_key(
                private_key.public_key()),
            critical=False
        ).add_extension(
            x509.ExtendedKeyUsage([extension]),
            critical=False,
        ).add_extension(
            x509.KeyUsage(
                digital_signature=True,
                content_commitment=False,
                key_encipherment=server,
                data_encipherment=False,
                key_agreement=False,
                key_cert_sign=False,
                crl_sign=False,
                encipher_only=False,
                decipher_only=False
            ),
            critical=False
        ).add_extension(
            x509.SubjectAlternativeName([
                x509.DNSName(cn),
            ]),
            critical=False,
        ).sign(
            private_key=private_key, algorithm=hashes.SHA256(),
            backend=default_backend()
        )

        with open(cert_path, "wb") as f:
            f.write(certificate.public_bytes(
                encoding=serialization.Encoding.PEM,
            ))

        return True

    def getCert(self, name):
        crt_path = os.path.join(self.key_path, "%s.crt" % name)

        if os.path.exists(crt_path):
            with open(crt_path, "rb") as crtfile:
                cert = x509.load_pem_x509_certificate(crtfile.read(),
                                                      default_backend())
            return cert
        else:
            return None

    def getCRL(self):
        now = datetime.datetime.utcnow()
        end = now + self.validity_time

        ca, private_key = self.getCA()

        crl = x509.CertificateRevocationListBuilder().issuer_name(
            ca.subject
        ).last_update(
            now
        ).next_update(
            end
        )

        if os.path.exists(self.crl_path):
            with open(self.crl_path, "rb") as crlfile:
                current_crl = x509.load_pem_x509_crl(crlfile.read(),
                                                     default_backend())

                for cert in current_crl:
                    crl = crl.add_revoked_certificate(cert)

        return crl

    def isRevoked(self, name):
        cert = self.getCert(name)
        if not cert:
            return None

        if os.path.exists(self.crl_path):
            with open(self.crl_path, "rb") as crlfile:
                current_crl = x509.load_pem_x509_crl(crlfile.read(),
                                                     default_backend())

                for revoked in current_crl:
                    if cert.serial_number == revoked.serial_number:
                        return True
        return False

    def revokeCertificate(self, name):
        if self.isRevoked(name):
            return None

        cert = self.getCert(name)
        if not cert:
            return None

        now = datetime.datetime.utcnow()
        crl = self.getCRL()
        ca, private_key = self.getCA()

        revoked = x509.RevokedCertificateBuilder().serial_number(
            cert.serial_number
        ).revocation_date(
            now
        ).build(default_backend())

        crl = crl.add_revoked_certificate(revoked)

        crl = crl.sign(
            private_key=private_key,
            algorithm=hashes.SHA256(),
            backend=default_backend()
        )

        with open(self.crl_path, "wb") as crlfile:
            crlfile.write(crl.public_bytes(
                encoding=serialization.Encoding.PEM
            ))

        return True

    def createCRL(self):
        self.createCSR('1')
        self.signCSR('1')
        return self.revokeCertificate('1')



from django.test import TestCase, Client
from django.urls import reverse, exceptions
from django.urls.resolvers import URLPattern
from django.contrib.auth.models import User, Group, Permission
from django.contrib.contenttypes.models import ContentType

from gamoto.tests import data

import os


# Create your tests here.
class TestRockFace(TestCase):
    def test_authentication_required(self):
        from gamoto import urls

        redirect_map = []
        for url in urls.urlpatterns:
            if not isinstance(url, URLPattern):
                continue
            if url.name == "login":
                continue
            try:
                urlr = reverse(url.name)
                redirect_map.append((url.name, urlr))
            except exceptions.NoReverseMatch:
                pass

        for name, path in redirect_map:
            r = self.client.get(reverse(name))
            self.assertRedirects(r, reverse('login')+'?next=' + path)

    def test_login(self):
        self.client.login(username='test', password='test')


class TestOpenvpn(TestCase):
    def setUp(self):
        self.sudo_commands = []
        ctype = ContentType.objects.create(
            app_label='subnet',
            model='group'
        )

        group = Group.objects.create(name="testgroup")

        user = User.objects.create(username="test", email="test@mail.com")
        user.groups.add(group)

        perm = Permission.objects.create(
            codename="network_10.1.2.0/24",
            name="My network",
            content_type=ctype
        )

        group.permissions.add(perm)

    def _fake_sudo(self, *a):
        self.sudo_commands.append(' '.join(a))
        result = b''
        if 'iptables-save' in a:
            result = bytes('\n'.join(data.IPTABLES_SAVE), encoding='ascii')

        elif a[1] == '-N':
            data.IPTABLES_SAVE.insert(5, ":%s - [0:0]" % a[2])

        elif a[1] == '-A':
            data.IPTABLES_SAVE.insert(6, ' '.join(a[1:]))

        return (result, b'')

    def _fake_get_user(self, name):
        return {
            'name': name,
            'uid': 1001,
            'gid': 1001,
            'gecos': name.capitalize,
            'home': '/var/lib/gamoto/users' + name,
            'shell': '/bin/false',
            'mfa': True
        }

    def test_connect(self):
        from gamoto.management.commands import openvpn
        from gamoto import users

        openvpn.users.sudo = self._fake_sudo
        openvpn.users.getUser = self._fake_get_user

        cmd = openvpn.Command()
        os.environ.setdefault('common_name', 'test')
        os.environ.setdefault('script_type', 'client-connect')
        os.environ.setdefault('ifconfig_pool_remote_ip', '10.88.20.35')

        cmd.handle()

        command_flow = [
            'iptables-save',
            '/sbin/iptables -N openvpn',
            '/sbin/iptables -I INPUT 1 -i tun0 -j openvpn',
            'iptables-save',
            '/sbin/iptables -A openvpn -i tun0 -s 10.88.20.35 -d 10.1.2.0/24'
            ' -m comment --comment "test"'
        ]

        self.assertListEqual(self.sudo_commands, command_flow)



from django.contrib.auth.models import Group, Permission, User
from django.contrib.contenttypes.models import ContentType
from django import forms
from django.core import validators


def validate_ipv4_cidr(value):
    """
    Validates a CIDR field
    """

    if '/' in value:
        value, prefix = value.split('/')
        try:
            assert(1 <= int(prefix) <= 32)
        except AssertionError:
            raise validators.ValidationError(
                "CIDR prefix must be between 1 and 32")
        except ValueError:
            raise validators.ValidationError(
                "CIDR prefix must be an integer")

    validators.validate_ipv4_address(value)


class CIDRField(forms.GenericIPAddressField):
    def __init__(self, *a, **kwargs):
        self.default_validators = [validate_ipv4_cidr]
        forms.CharField.__init__(self, *a, **kwargs)


class GroupForm(forms.ModelForm):
    """
    Form for creating endpoint groups
    """
    class Meta:
        model = Group
        exclude = ()

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        permission_query = Permission.objects.filter(
            content_type=ContentType.objects.get(app_label='subnet')
        ).order_by('name')

        self.fields['permissions'].queryset = permission_query
        self.fields['permissions'].label_from_instance = self.label_permission

    def label_permission(self, obj):
        subnet = obj.codename.split('_')[-1]
        return obj.name + ': ' + subnet


class UserGroupForm(forms.ModelForm):
    class Meta:
        model = User
        exclude = ('password', 'user_permissions', 'username', 'first_name',
                   'last_name', 'last_login', 'is_staff', 'is_active',
                   'is_superuser', 'email', 'date_joined')


class SubnetForm(forms.ModelForm):
    """
    Form for creating endpoint groups
    """
    class Meta:
        model = Permission
        exclude = ()


class GroupSubnetForm(forms.Form):
    """
    Add subnet to group form
    """
    name = forms.CharField(min_length=2, max_length=128)
    subnet = CIDRField()



IPTABLES_SAVE = [
    "# Generated by iptables-save v1.6.0 on Fri May 10 10:17:18 2019",
    "*filter",
    ":INPUT ACCEPT [27:6332]",
    ":FORWARD ACCEPT [0:0]",
    ":OUTPUT ACCEPT [1:40]",
    "COMMIT",
    "# Completed on Fri May 10 10:17:18 2019\n"
]



# Creates and manage users
from django.conf import settings

from gamoto import ca

from io import BytesIO

import pyotp

import subprocess
import random
import pwd
import os
import shutil
import zipfile


def getSystemUID():
    return pwd.getpwnam(settings.GAMOTO_USER).pw_uid


def sudo(*a):
    proc = subprocess.Popen(['sudo'] + list(a), stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    return proc.communicate()


def sudoWrite(filename, data):
    proc = subprocess.Popen(
        ['sudo', 'tee', filename],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )

    r = proc.communicate(bytes(data, encoding='ascii'))


def sudoRead(filename):
    proc = subprocess.Popen(
        ['sudo', 'cat', filename],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )

    data, err = proc.communicate()

    if err:
        raise Exception(err)

    return data


def createVPN(name):
    myca = ca.CertificateAuthority(
        settings.CA_SETUP['ou'],
        settings.CA_SETUP['org'],
        settings.CA_SETUP['email'],
        settings.CA_SETUP['country'],
        settings.CA_SETUP['state'],
        settings.CA_SETUP['city'],
        key_path=settings.CA_PATH
    )

    myca.createCSR(name)
    myca.signCSR(name)


def getVPNZIP(name):
    user_cert = os.path.join(settings.CA_PATH, name + '.crt')
    user_key = os.path.join(settings.CA_PATH, name + '.key')
    server_cert = os.path.join(settings.CA_PATH, 'ca.crt')

    if not os.path.exists(user_cert):
        raise Exception("Certificate does not exist")

    config = [
        "client",
        "dev tun",
        "proto udp",
        "remote %s %s" % (settings.OPENVPN_HOSTNAME, settings.OPENVPN_PORT),
        "resolv-retry infinite",
        "nobind",
        "persist-key",
        "persist-tun",
        "ca ca.crt",
        "cert %s.crt" % name,
        "key %s.key" % name,
        "verb 3",
        "auth-user-pass"
    ]

    zip_io = BytesIO()

    with zipfile.ZipFile(zip_io, mode="w") as vpnzip:
        vpnzip.write(str(user_cert), arcname=name + '.crt')
        vpnzip.write(str(user_key), arcname=name + '.key')
        vpnzip.write(str(server_cert), arcname='ca.crt')

        vpnzip.writestr('client.ovpn', '\n'.join(config))

    return zip_io.getvalue()


def getUser(name):
    """
    Check if a user exists and return all the info about them
    as a dict
    """
    try:
        passwd = pwd.getpwnam(name)
    except KeyError:
        return None

    stdout, stderr = sudo(
        'stat', os.path.join(passwd.pw_dir, '.google_authenticator')
    )

    if stdout:
        mfa = True
    else:
        mfa = False

    return {
        'name': passwd.pw_name,
        'uid': passwd.pw_uid,
        'gid': passwd.pw_gid,
        'gecos': passwd.pw_gecos,
        'home': passwd.pw_dir,
        'shell': passwd.pw_shell,
        'mfa': mfa
    }


def createUser(name):
    passwd = getUser(name)

    if not passwd:
        sudo(
            '/usr/sbin/useradd',
            '-s', '/bin/false',
            '-b', settings.USER_PATH,
            '-g', settings.GAMOTO_GROUP,
            '-N', '-M', name
        )
        sudo('mkdir', os.path.join(settings.USER_PATH, name))
        sudo(
            'chown',
            '%s:%s' % (name, settings.GAMOTO_GROUP),
            os.path.join(settings.USER_PATH, name)
        )


def configureTOTP(name):
    google_auth = os.path.join(settings.USER_PATH, name,
                               '.google_authenticator')
    key = pyotp.random_base32()
    totp = pyotp.totp.TOTP(key)

    codes = [str(random.randint(10000000, 99999999)) for i in range(5)]
    authfi = '\n'.join([key, '" RATE_LIMIT 3 30', '" TOTP_AUTH'] + codes)

    sudoWrite(google_auth, authfi + '\n')

    sudo('chown', '%s:%s' % (name, settings.GAMOTO_GROUP), google_auth)
    sudo('chmod', '0600', google_auth)

    uri = totp.provisioning_uri(name, issuer_name=settings.CA_SETUP['org'])
    return (codes, uri)



from setuptools import setup, find_packages

setup(
    name='gamoto',
    version='0.0.4',
    url='http://github.com/calston/gamoto',
    description='A web portal for VPN self service',
    classifiers=[
        "Programming Language :: Python",
    ],
    author='Colin Alston',
    author_email='colin.alston@gmail.com',
    license='BSD',
    packages=find_packages(),
    scripts=['bin/gamoto'],
    include_package_data=True,
    install_requires=[
        'pyopenssl',
        'django==2.2',
        'social-auth-app-django',
        'pyotp',
        'gunicorn'
    ],
)
